module tools_io
  implicit none

  private
  public :: tictac
  public :: getargs
  public :: error
  public :: header
  public :: zatguess, nameguess
  public :: upper
  public :: realloc
  public :: isinteger

contains

  !> Outputs a tictac: date and time.
  function ffdate()
    use param
    character*(mline) :: ffdate

    integer :: values(8)

    call date_and_time(values=values)
    write (ffdate,'(I4.4,".",I2.2,".",I2.2,", ",&
       &I2.2,":",I2.2,":",I2.2,".",I3.3)') values(1:3),values(5:8)

  end function ffdate

  !> Interface to the timer routines of different computers. Returns
  !> machine seconds at the calling time.
  subroutine tictac(mesg)
    use param
    character*(*), intent(in) :: mesg !< Prefix message

    character*(mline) :: sdate

    sdate = ffdate()
    write (uout,'(A," -- ",A/)') trim(mesg), trim(adjustl(sdate))

  end subroutine tictac

  !> Get command line arguments and argument count
  !> on return, argc will contain the number of arguments and
  !> argv will be a character array containing the arguments.
  subroutine getargs (argc, argv)
    use param

    integer, intent(out) :: argc
    character*(*), intent(out) :: argv(2)
    
    character*(mline) :: line  !local string for parsing
    integer :: length         !total length of command line

    argc=0
    length = 1
    do while (length > 0)
       call getarg (argc+1,line)
       length=len(trim(adjustl(line)))
       if (length.gt.0) then
          argc=argc+1
          if (argc > 2) call error('getargs','Usage: nciplot [file.in [file.out]]',faterr)
          argv(argc)=line(1:length)
       endif
    end do

  end subroutine getargs

  !> Send an error message 'message' to stdout, coming from routine
  !> 'routine'. errortype is the error code (see mod_param.f90).
  subroutine error (routine,message,errortype)
    use param

    character*(*), intent(in) :: routine !< routine calling the error
    character*(*), intent(in) :: message !< the message
    integer, intent(in) :: errortype !< fatal, warning or info

    character*(20)      chtype

    ! message styles shamelessly copied from abinit.
    if (errortype.eq.faterr) then
       chtype='ERROR'
    else if (errortype.eq.warning) then
       chtype='WARNING'
    else if (errortype.eq.noerr) then
       chtype='COMMENT'
    else
       chtype='UNKNOWN'
    endif
    write (uout,100) trim(adjustl(chtype)),trim(adjustl(routine)),&
       trim(adjustl(message))
    if (errortype.eq.faterr) then
       write (stderr,100) trim(adjustl(chtype)),trim(adjustl(routine)),&
          trim(adjustl(message))
       if (uin /= stdin) close(uin)
       if (uout /= stdout) close(uout)
       stop 1
    else if(errortype.eq.warning) then
       nwarns = nwarns + 1
    else if(errortype.eq.noerr) then
       ncomms = ncomms + 1
    endif

100 format (A,"(",A,"): ",A)

  end subroutine error

  subroutine header()
    use param

    write (uout,1000)
1000 format(' # ----------------- NCIPLOT ------------------------',&
      /,    ' # --- PLOTTING NON COVALENT INTERACTION REGIONS ----',/,&
            ' # ---             E.R. Johnson                  ----',/,& 
            ' # ---          J. Contreras-Garcia              ----',/,&
            ' # ----------    Duke University         ------------',/,&
            ' # --------------------------------------------------',/,&
            ' # ---              Please cite                  ----',/,&
            ' # --J. Am. Chem. Soc., 2010, 132 (18), pp 6498–6506-',/,&
            ' # --------------------------------------------------',///&
            ' # --------------------------------------------------',/,&
            ' # ---     Contributions for the wfn properties  ----',/,&
            ' # ---      from H. L. Schmider are acknowledged  ---',/,&
            ' # --------------------------------------------------',/&
            ' #')
    
  end subroutine header

  !> Guess the atomic number from the atomic name. A value
  !> of -1 is returned if the atom is not recognized.
  !> A valid atomic name is made of:
  !>   1) one or two capital or small case letters corresponding to
  !>      a legal atomic symbol
  !>   2) any other additional characters with the only excepcion that
  !>      the first character following (1) should not be a letter
  function zatguess (atname)
    
    integer :: zatguess !< Ouptut atomic number
    character*(*), intent(in) :: atname !< Input atomic symbol (case insensitive)

    character*(2) :: atsymbol
    integer       :: i, idx
    character*(2) :: an(1:103)

    data (an(i), i = 1, 103) &
         & /'H ', 'HE', 'LI', 'BE', 'B ', 'C ', 'N ', 'O ', 'F ', 'NE',&
         &  'NA', 'MG', 'AL', 'SI', 'P ', 'S ', 'CL', 'AR', 'K ', 'CA',&
         &  'SC', 'TI', 'V ', 'CR', 'MN', 'FE', 'CO', 'NI', 'CU', 'ZN',&
         &  'GA', 'GE', 'AS', 'SE', 'BR', 'KR', 'RB', 'SR', 'Y ', 'ZR',&
         &  'NB', 'MO', 'TC', 'RU', 'RH', 'PD', 'AG', 'CD', 'IN', 'SN',&
         &  'SB', 'TE', 'I ', 'XE', 'CS', 'BA', 'LA', 'CE', 'PR', 'ND',&
         &  'PM', 'SM', 'EU', 'GD', 'TB', 'DY', 'HO', 'ER', 'TM', 'YB',&
         &  'LU', 'HF', 'TA', 'W ', 'RE', 'OS', 'IR', 'PT', 'AU', 'HG',&
         &  'TL', 'PB', 'BI', 'PO', 'AT', 'RN', 'FR', 'RA', 'AC', 'TH',&
         &  'PA', 'U ', 'NP', 'PU', 'AM', 'CM', 'BK', 'CF', 'ES', 'FM',&
         &  'MD', 'NO', 'LW'/

    idx = 1
    do while(atname(idx:idx) == ' ' .and. idx < len(atname)-1)
       idx = idx + 1
    end do
    if (len(trim(adjustl(atname))) > 1) then
       atsymbol = atname(idx:idx+1)
    else
       atsymbol = atname(idx:idx)
    end if
    call upper(atsymbol)
    if (atsymbol(2:2) .lt. 'A' .or. atsymbol(2:2) .gt. 'Z') atsymbol(2:2) = ' '
    do i = 1, 103
       if (atsymbol .eq. an(i)) then
          zatguess = i
          return
       endif
    enddo
    zatguess = -1
  end function zatguess

  !> The reverse of zatguess. Given the atomic number, return the 
  !> chemical symbol of the atom, or XX if it is not recognized.
  function nameguess (zat)
    
    integer, intent(in) :: zat !< Input atomic number
    character*(2) :: nameguess !< Output atomic symbol 

    character*(2) :: an(1:103)
    integer :: i

    data (an(i), i = 1, 103) &
         & /'h_', 'he', 'li', 'be', 'b_', 'c_', 'n_', 'o_', 'f_', 'ne',&
         &  'na', 'mg', 'al', 'si', 'p_', 's_', 'cl', 'ar', 'k_', 'ca',&
         &  'sc', 'ti', 'v_', 'cr', 'mn', 'fe', 'co', 'ni', 'cu', 'zn',&
         &  'ga', 'ge', 'as', 'se', 'br', 'kr', 'rb', 'sr', 'y_', 'zr',&
         &  'nb', 'mo', 'tc', 'ru', 'rh', 'pd', 'ag', 'cd', 'in', 'sn',&
         &  'sb', 'te', 'i_', 'xe', 'cs', 'ba', 'la', 'ce', 'pr', 'nd',&
         &  'pm', 'sm', 'eu', 'gd', 'tb', 'dy', 'ho', 'er', 'tm', 'yb',&
         &  'lu', 'hf', 'ta', 'w_', 're', 'os', 'ir', 'pt', 'au', 'hg',&
         &  'tl', 'pb', 'bi', 'po', 'at', 'rn', 'fr', 'ra', 'ac', 'th',&
         &  'pa', 'u_', 'np', 'pu', 'am', 'cm', 'bk', 'cf', 'es', 'fm',&
         &  'md', 'no', 'lw'/

    if (zat > 0 .and. zat < 104) then
       nameguess = an(zat)
    else
       nameguess = 'xx'
    end if
    
  end function nameguess

  !> In-place convert string to uppercase.
  subroutine upper(string)
    
    character*(*), intent(inout) :: string !< Input string, same as the function result on output

    integer :: i

    do i = 1, len(string)
       if (string(i:i).ge.'a' .and. string(i:i).le.'z') &
          string(i:i) = char(ichar(string(i:i)) + ichar('A') - ichar('a'))
    enddo

  end subroutine upper

  !> Adapt the size of an allocatable 1D real*8 array
  subroutine realloc(a,nnew)
    use param

    real*8, intent(inout), allocatable :: a(:) !< Input array, real*8, 1D
    integer, intent(in) :: nnew !< new dimension
    
    real*8, allocatable :: temp(:)
    integer :: nold
    
    if (.not.allocated(a)) &
       call error('realloc','array not allocated',faterr)
    nold = size(a)
    allocate(temp(nnew))
    
    temp(1:min(nnew,nold)) = a(1:min(nnew,nold))
    call move_alloc(temp,a)

  end subroutine realloc

  !> Get integer value from input text. If a valid integer is not
  !> found, then return .false.
  logical function isinteger (ival,line,lp)
    use param
    
    character*(mline), intent(in) :: line !< Input string
    integer, intent(inout) :: lp !< Pointer to current position on string
    integer, intent(out) :: ival !< Integer value read, same as function result

    integer :: i

    do while (line(lp:lp) .eq. " ")
       lp=lp+1
    enddo
    i=lp
    if (line(i:i) .eq. '+' .or. line(i:i) .eq. '-') i=i+1
    if (isdigit(line(i:i))) then 
       do while (isdigit(line(i:i)))
          i=i+1
       enddo
       if (line(i:i) .eq. " " .or. line(i:i) .eq. char(0) .or. line(i:i).eq.char(10)) then
          ival=atoi(line(lp:i-1))
          lp = i
          isinteger=.true.
       else
          ival=0
          isinteger=.false.
       endif
    else
       ival=0
       isinteger=.false.
    endif

    return
  end function isinteger

  !> Return true if c is a digit (private).
  logical function isdigit (c)
    
    character*(1), intent(in) :: c !< Is the character c a digit?

    isdigit = c.ge.'0' .and. c.le.'9'

  end function isdigit

  !> Convert ascii string to integer (private).
  integer function atoi (string)
    
    character*(*), intent(in) :: string !< Input string

    integer           i, sign

    atoi=0
    i=1
    do while (string(i:i).eq." ")
       i=i+1
    end do
    sign=1
    if(string(i:i) .eq. '+' .or. string(i:i) .eq. '-') then 
       if (string(i:i) .eq. '-') sign=-1
       i=i+1
    endif
    do while (isdigit(string(i:i)))
       atoi=10*atoi+ichar(string(i:i))-ichar('0')
       i=i+1
       if (i > len(string)) then
          exit
       end if
    end do
    atoi=atoi*sign
    return

  end function atoi

end module tools_io
